/*
 * ClassTable.java - part of the CodeAid plugin.
 * Copyright (C) 1999 Jason Ginchereau
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package codeaid.info;

import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import java.util.Enumeration;
// Collections API
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.SortedMap;
import java.util.TreeMap;

import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import java.lang.reflect.Modifier;


/**
 * Maps class names to information that is known about those classes.
 * Information about each class is stored in a <code>ClassInfo</code> object.
 * All methods are thread-safe.
 *
 * @see ClassInfo
 *
 * @author Jason Ginchereau
 */
public final class ClassTable implements java.io.Serializable {
    public static final String UNNAMED_PACKAGE = "<unnamed package>";

    private static Map staticInstanceTable = new HashMap();
    private static final Object DEFAULT_KEY = new String();


    /**
     * Gets a static instance of a <code>ClassTable</code> that can
     * be shared by everyone with the same key.
    **/
    public static synchronized ClassTable getInstance(Object key) {
        ClassTable instance;
        if ((instance = (ClassTable) staticInstanceTable.get(key)) == null) {
            staticInstanceTable.put(key, instance = new ClassTable());
        }
        return instance;
    }


    /**
     * Gets a static instance of a <code>ClassTable</code> using
     * the default key.
    **/
    public static ClassTable getInstance() {
        return getInstance(DEFAULT_KEY);
    }


    private transient ClassTable parent;
    private transient SortedMap  arrays;
    private SortedMap packages;
    private SortedMap classes;


    /**
     * Creates a new <code>ClassTable</code>.
    **/
    public ClassTable() {
        this(null);
    }


    /**
     * Creates a new <code>ClassTable</code> which will delegate to a
     * parent <code>ClassTable</code> for unrecognized classes.
    **/
    public ClassTable(ClassTable parent) {
        this.parent = parent;
        arrays = new TreeMap(new StringComparator());
        packages = new TreeMap(new StringComparator());
        classes = new TreeMap(new StringComparator());
    }


    /**
     * Gets the parent <code>ClassTable</code>, which is be queried for
     * classes that are not recognized by this <code>ClassTable</code>.
    **/
    public ClassTable getParent() {
        return parent;
    }


    /**
     * Sets the parent <code>ClassTable</code>, which will be queried for
     * classes that are not recognized by this <code>ClassTable</code>.
    **/
    public synchronized void setParent(ClassTable parent) {
        this.parent = parent;
    }


    /**
     * Gets a list of all packages in the database.
     *
     * @return a sorted <code>List</code> of <code>Strings</code>, where
     *         each <code>String</code> is a name of a package in the database.
     *         This <code>List</code> may be safely modified by the user, as
     *         it is no longer connected to the database after it is returned.
    **/
    public synchronized List getPackages() {
        return new LinkedList(packages.keySet());
    }


    /**
     * Gets a list of all known classes in a package.
     *
     * @param  packageName the name of the package, using the java . notation.
     * @return a sorted <code>List</code> of <code>String</code> names,
     *         or <code>null</code> if the package is unknown.  This
     *         <code>List</code> may be safely modified by the user, as
     *         it is no longer connected to the database after it is returned.
    **/
    public synchronized List getClassesInPackage(String packageName) {
        Map pmap = (Map) packages.get(packageName);
        if (pmap != null) {
            return new LinkedList(pmap.keySet());
        } else if (parent != null) {
            return parent.getClassesInPackage(packageName);
        } else {
            return null;
        }
    }


    /**
     * Gets information about a class, given a class name.
     *
     * @param  fullClassName a fully-qualified class name.  Note that for
     *         inner classes, the <code>$</code> notation is used in the
     *         class name.  Example: <code>java.lang.Character$Subset</code>
     * @return the <code>ClassInfo</code> object for the class, or
     *         <code>null</code> if the class is not in the database.
    **/
    public synchronized ClassInfo get(String fullClassName) {
        if (fullClassName.endsWith("[]")) {
            ClassInfo aci = (ClassInfo) arrays.get(fullClassName);

            if (aci == null) {
                String originalName = fullClassName;
                int dimensions = 0;
                do {
                    fullClassName = fullClassName.substring(0, fullClassName.length()-2);
                    dimensions++;
                } while (fullClassName.endsWith("[]"));

                ClassInfo ci = get(fullClassName);
                if (ci != null) {
                    aci = new ArrayClassInfo(ci, dimensions);
                    arrays.put(originalName, aci);
                }
            }
            return aci;
        } else {
            Object o = classes.get(fullClassName);
            if (o instanceof ZipFile) {
                load(fullClassName);
                o = classes.get(fullClassName);
            }

            if (o != null) {
                return (ClassInfo) o;
            } else if (parent != null) {
                return parent.get(fullClassName);
            } else {
                return null;
            }
        }
    }


    /**
     * Enters class information into the database.
     *
     * @param  classInfo a <code>ClassInfo</code> object that information about
     *         a class to be put in the database.  This object should not be
     *         modified after it is put in the database.
    **/
    public synchronized void put(ClassInfo classInfo) {
        if (classInfo instanceof ArrayClassInfo) {
            throw new IllegalArgumentException("Can't put array classes in ClassTable.");
        }

        // Don't put inner classes in the package map.
        if (classInfo.getDeclaringClass() == null) {
            String packageName = classInfo.getPackage();
            if (packageName == null) { packageName = UNNAMED_PACKAGE; }
            Map pmap = (Map) packages.get(packageName);
            if (pmap == null) {
                packages.put(packageName, pmap = new TreeMap(new StringComparator()));
            }
            pmap.put(classInfo.getName(), classInfo);
        }

        classes.put(classInfo.getFullName(), classInfo);
    }


    /**
     * Removes a class from the database.
     *
     * @param  fullClassName a fully-qualified class name.  Note that for
     *         inner classes, the <code>$</code> notation is used in the
     *         class name.
     * @return the <code>ClassInfo</code> object that was removed from the
     *         database, or <code>null</code> if there was not one.
    **/
    public synchronized ClassInfo remove(String fullClassName) {
        ClassInfo ci = (ClassInfo) classes.remove(fullClassName);
        if (ci != null) {
            String packageName = ci.getPackage();
            if (packageName == null) { packageName = UNNAMED_PACKAGE; }
            Map pmap = (Map) packages.get(packageName);
            if (pmap != null) {
                pmap.remove(ci.getName());
                if (pmap.isEmpty()) { packages.remove(packageName); }
            }
        }
        return ci;
    }


    /**
     * Tests if a class is contained in the database.
     *
     * @param  fullClassName a fully-qualified class name.  Note that for
     *         inner classes, the <code>$</code> notation is used in the
     *         class name.
     * @return <code>true</code> if information about the class is in the
     *         database.
    **/
    public synchronized boolean contains(String fullClassName) {
        return classes.containsKey(fullClassName);
    }


    public boolean isAssignableTo(String fromClass, String toClass,
                                  boolean defaultAnswer
    ) {
        ClassInfo fci = get(fromClass);
        ClassInfo tci = get(toClass);

        // return defaultAnswer when one or both classes are unrecognized
        if (fci == null || tci == null) { return defaultAnswer; }

        // a class is assignable to itself
        if (fci.equals(tci)) { return true; }

        // handling arrays
        if (fci instanceof ArrayClassInfo && tci instanceof ArrayClassInfo) {
            ArrayClassInfo afci = (ArrayClassInfo)fci;
            ArrayClassInfo atci = (ArrayClassInfo)tci;
            if (afci.getDimensions() != atci.getDimensions()) {
                return false;
            }
            return isAssignableTo(afci.getArrayType().getFullName(),
                atci.getArrayType().getFullName(), defaultAnswer);
        }

        // a class is assignable if its superclass is assignable
        if (fci.getSuperclass() != null &&
            isAssignableTo(fci.getSuperclass(), toClass, defaultAnswer)
        ) {
            return true;
        }

        // a class is assignable if any of its interfaces are assignable
        String[] ifs = fci.getInterfaces();
        for (int i = 0; i < ifs.length; i++) {
            if (isAssignableTo(ifs[i], toClass, defaultAnswer)) {
                return true;
            }
        }

        return false;
    }


    public boolean isAccessAllowed(ClassInfo in, MemberInfo target,
                                   boolean requireStatic, boolean defaultAnswer
    ) {
        int mods = target.getModifiers();
        if (requireStatic && (mods & Modifier.STATIC) == 0) { return false; }

        String dc = target.getDeclaringClass();
        ClassInfo dci = (dc != null? (ClassInfo) get(dc) : null);
        if (dci != null) {
            if(!isAccessAllowed(in, dci, false, defaultAnswer)) { return false; }
        } else if(dc != null) {
            if (!defaultAnswer) { return false; }
        }

        if (target instanceof ClassInfo && dc == null) { // top-level class
            if ((mods & Modifier.PUBLIC) != 0) {
                return true;
            } else {
                String ip = (in != null? in.getPackage() : null);
                String cp = ((ClassInfo)target).getPackage();
                if ((ip == null && cp == null) ||
                    (ip != null && cp != null && ip.equals(cp))
                ) {
                    return true;
                }
            }
            return false;
        }

        if ((mods & Modifier.PUBLIC) != 0) { // public access
            return true;
        }

        if (in != null && dci != null) {
            String cn = dci.getFullName();
            String cp = dci.getPackage();
            String icn = in.getFullName();
            String icp = in.getPackage();

            if ((mods & Modifier.PROTECTED) != 0) { // protected access
                if (isAssignableTo(icn, cn, false) ||
                    (cp == null && icp == null) ||
                    (cp != null && icp != null && cp.equals(icp))
                ) {
                    return true;
                }
            }
            else if ((mods & Modifier.PRIVATE) == 0) { // package access
                if ((cp == null && icp == null) ||
                    (cp != null && icp != null && cp.equals(icp))
                ) {
                    return true;
                }
            }
            else { // private access
                if (cn.equals(icn)) { return true; }
            }

            // Allow inner classes to access anything that their declaring
            // classes can access.. is this correct?
            String idc = in.getDeclaringClass();
            if (idc != null) {
                ClassInfo idci = (ClassInfo) get(idc);
                if (idci != null) {
                    return isAccessAllowed(idci, target, requireStatic, defaultAnswer);
                }
            }
        }

        return defaultAnswer;
    }


    /**
     * Gets the size of the database.
     *
     * @return the number of classes in the database.
    **/
    public synchronized int size() {
        return classes.size();
    }


    private void load(String name) {
        ZipFile file = (ZipFile) classes.get(name);
        ZipEntry entry = file.getEntry(name);
        ClassInfo ci = null;
        try {
            ObjectInputStream in = new ObjectInputStream(file.getInputStream(entry));
            ci = (ClassInfo) in.readObject();
            in.close();
        } catch(java.io.InvalidClassException ex) {
            System.err.println("" + file.getName() + " is an outdated version.");
            System.err.println("Please create or download an updated version.");
            // Remove all entries of this file so the error message is not repeated.
            Collection values = classes.values();
            while (values.remove(file)) {}
        } catch(Exception ex) {
            System.err.println("Failed to load ClassInfo: " + name);
            System.err.println(ex.toString());
            // Remove this entry so the error message is not repeated.
            classes.remove(name);
        }
        if (ci != null) { put(ci); }
    }


    /**
     * Loads <code>ClassInfo</code> entry information out of a
     * <code>ZipFile</code>.  The entries themselves will not actually be
     * loaded into memory until they are needed.  Therefore, the
     * <code>ZipFile</code> should not be closed after loading.
    **/
    public synchronized void preLoad(ZipFile file) {
        for(Enumeration e = file.entries(); e.hasMoreElements(); ) {
            ZipEntry entry = (ZipEntry) e.nextElement();
            String name = entry.getName();
            if (name.lastIndexOf('$') < 0) {
                String packageName = UNNAMED_PACKAGE;
                int i = name.lastIndexOf('.');
                if (i >= 0) { packageName = name.substring(0, i); }
                Map pmap = (Map) packages.get(packageName);
                if (pmap == null) {
                    packages.put(packageName, pmap = new TreeMap(new StringComparator()));
                }
                pmap.put((i >= 0? name.substring(i+1) : name), file);
            }
            classes.put(name, file);
        }
    }


    public synchronized void unload(ZipFile file) {
        for (Enumeration e = file.entries(); e.hasMoreElements();) {
            ZipEntry entry = (ZipEntry) e.nextElement();
            String name = entry.getName();
            if (name.lastIndexOf('$') < 0) {
                String packageName = UNNAMED_PACKAGE;
                int i = name.lastIndexOf('.');
                if (i >= 0) { packageName = name.substring(0, i); }
                Map pmap = (Map)packages.get(packageName);
                if (pmap != null) {
                    String s = (i >= 0 ? name.substring(i+1) : name);
                    Object o = pmap.get(s);
                    if (o == file) {
                        pmap.remove(s);
                    }
                }
            }
            Object o = classes.get(name);
            if (o == file) {
                classes.remove(name);
            }
        }
    }


    /**
     * Loads <code>ClassInfo</code> entry information out of a
     * <code>ZipFile</code>.  The entries themselves will not actually be
     * loaded into memory until they are needed.  Therefore, the
     * <code>ZipFile</code> should not be closed after loading.
    **/

    /**
     * If any <code>ClassInfo</code> entries have not yet been loaded from
     * their <code>ZipFile</code>, they are loaded now.
     * <p>
     * Calling <code>preLoad(ZipFile)</code>, then <code>loadAll()</code>
     * Has the same result as calling <code>read(ZipInputStream)</code>.
    **/
    public synchronized void loadAll() throws IOException {
        for(Iterator i = classes.entrySet().iterator(); i.hasNext(); ) {
            Map.Entry entry = (Map.Entry) i.next();
            if (entry.getValue() instanceof ZipFile) {
                load((String) entry.getKey());
            }
        }
    }


    /**
     * Loads <code>ClassInfo</code> entries from a <code>ZipInputStream</code>.
     * All entries will be stored in memory.
    **/
    public synchronized void read(ZipInputStream in) throws IOException {
        ClassInfo ci;
        ZipEntry entry = null;
        try {
            while ((entry = in.getNextEntry()) != null) {
                ObjectInputStream objIn = new ObjectInputStream(in);
                ci = (ClassInfo) objIn.readObject();
                put(ci);
                in.closeEntry();
            }
        } catch(EOFException eofex) {
        } catch(Exception ex) {
            System.err.println("Failed to read ClassInfo: " +
                               (entry != null? entry.getName() : null));
            System.err.println(ex.toString());
        }
    }


    /**
     * Writes <code>ClassInfo</code> entries out to a
     * <code>ZipOutputStream</code>.
    **/
    public synchronized void write(ZipOutputStream out) throws IOException {
        for (Iterator i = classes.entrySet().iterator(); i.hasNext(); ) {
            Map.Entry entry = (Map.Entry) i.next();
            if (entry.getValue() instanceof ClassInfo) {
                out.putNextEntry(new ZipEntry((String) entry.getKey()));
                ObjectOutputStream objOut = new ObjectOutputStream(out);
                objOut.writeObject(entry.getValue());
                out.closeEntry();
            }
        }
        out.finish();
    }
}

