/*
 * DocumentReader.java - a Reader that reads from a javax.swing.text.Document.
 * Copyright (C) 1999 Jason Ginchereau
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package codeaid.parser;

import java.io.IOException;
import java.io.Reader;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.Position;
import javax.swing.text.Segment;


/**
 * Class to provide Reader functionality from a portion of a
 * Document.
 */
final class DocumentReader extends Reader {
    private Document doc;
    private Segment  segment;
    private int      p0;    // start position
    private int      p1;    // end position, or Integer.MAX_VALUE
    private Position pos;   // pos in document
    private int      index; // index into array of the segment


    public DocumentReader(Document doc) {
        this.doc = doc;
        this.segment = new Segment();
        try { reset(0, Integer.MAX_VALUE); } catch(BadLocationException blex) { }
    }


    public DocumentReader(Document doc, int p0, int p1)
        throws BadLocationException
    {
        this.doc = doc;
        this.segment = new Segment();
        reset(p0, p1);
    }


    public void reset() {
        try { reset(0, Integer.MAX_VALUE); } catch(BadLocationException blex) { }
    }


    // Avoids having to re-create lots of DocumentReader instances.
    public void reset(int p0, int p1) throws BadLocationException {
        this.p0 = p0;
        this.p1 = p1;
        pos = doc.createPosition(p0);
        index = 0;
        segment.offset = segment.count = 0;
    }


    public int read() throws IOException {
        if (index >= segment.offset + segment.count) {
            if (pos.getOffset() >= p1) { return -1; }
            loadSegment();
            if (segment.count == 0) { return -1; }
        }
        return segment.array[index++];
    }


    public int read(char[] cbuf, int off, int len) throws IOException {
        int count = 0;
        int copyCount = 0;
        while (count < len) {
            if (index >= segment.offset + segment.count) {
                if (pos.getOffset() >= p1) { break; }
                loadSegment();
                if (segment.count == 0) { break; }
            }
            copyCount = Math.min(len - count, segment.offset + segment.count - index);
            System.arraycopy(segment.array, index, cbuf, count, copyCount);
            count += copyCount;
            index += copyCount;
        }
        return (count > 0) ? count : -1;
    }


    private void loadSegment() throws IOException {
        synchronized (doc) {
            int offset = pos.getOffset();
            int n = Math.min(1024, Math.min(p1, doc.getLength()) - offset);
            try {
                doc.getText(offset, n, segment);
                pos = doc.createPosition(offset + n);
            } catch (BadLocationException e) {
                throw new IOException("Bad location");
            }
        }
        index = segment.offset;
    }


    public boolean ready() throws IOException {
        return index < segment.offset + segment.count;
    }


    public void close() throws IOException {
        // Do nothing.
        // This technically violates the spec for java.io.Reader,
        // but this method should never be called anyway.
    }
}

