/*
 * SymbolTable.java - part of the CodeAid plugin for jEdit
 * Copyright (C) 1999 Jason Ginchereau
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package codeaid.parser;

import codeaid.info.ClassInfo;
import codeaid.info.MemberInfo;

// Collections API
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;


/**
 * A table of all symbols declared within a source file.
 * Each symbol is represented by an object of some subclass of
 * <code>MemberInfo</code>.
 *
 * @author Jason Ginchereau
 */
public class SymbolTable {
    private SymbolTable superTable;
    private SymbolTable subTables;
    private SymbolTable next;
    private MemberInfo  symbol;
    private List        symbols;
    private int         beginLine;
    private int         beginColumn;
    private int         endLine;
    private int         endColumn;


    SymbolTable(SymbolTable superTable) {
        this.superTable = superTable;
        this.subTables = null;
        this.next = null;
        this.symbol = null;
        this.symbols = new LinkedList();
        this.beginLine   = Integer.MAX_VALUE;
        this.beginColumn = Integer.MAX_VALUE;
        this.endLine     = Integer.MAX_VALUE;
        this.endColumn   = Integer.MAX_VALUE;
    }


    void beginBuild(int line, int col) {
        if (beginLine != Integer.MAX_VALUE) {
            if (endLine == Integer.MAX_VALUE) {
                throw new IllegalStateException("Build already begun.");
            } else {
                throw new IllegalStateException("Build already ended.");
            }
        }
        beginLine = line;
        beginColumn = col;
        if (superTable != null) { superTable.addSubTable(this); }
    }


    void endBuild(int line, int col) {
        if (beginLine == Integer.MAX_VALUE) {
            throw new IllegalStateException("Build not begun.");
        }
        if (endLine != Integer.MAX_VALUE) {
            throw new IllegalStateException("Build already ended.");
        }
        endLine = line;
        endColumn = col;
    }


    void addSymbol(MemberInfo sym) {
        symbols.add(sym);
        //System.out.println("added symbol: " + sym);
        //System.out.println(this.toString());
    }


    void setSymbol(MemberInfo mi) {
        this.symbol = mi;
    }


    private void addSubTable(SymbolTable st) {
        st.next = null;
        if (subTables == null) {
            subTables = st;
        } else {
            SymbolTable t;
            for (t = subTables; t.next != null; t = t.next) {}
            t.next = st;
        }
        //System.out.println("added subTable");
        //System.out.println(this.toString());
    }


    public SymbolTable getSuperTable() {
        return superTable;
    }


    public SymbolTable getSubTable(int line, int col) {
        for (SymbolTable t = subTables; t != null; t = t.next) {
            if (t.contains(line, col)) { return t; }
        }
        return null;
    }


    public boolean contains(int line, int col) {
        if (beginLine == line) {
            return col >= beginColumn;
        } else if (beginLine < line && line < endLine) {
            return true;
        } else if (endLine == line) {
            return col < endColumn;
        } else {
            return false;
        }
    }


    public synchronized List getSymbolPath(int line, int col) {
        List sp = new LinkedList();
        if (contains(line, col)) { findSymbolPath(line, col, sp); }
        return sp;
    }


    private void findSymbolPath(int line, int col, List path) {
        ClassInfo ci = null;
        for(SymbolTable t = subTables; t != null; t = t.next) {
            if (t.contains(line, col)) {
                t.findSymbolPath(line, col, path);
                break;
            }
            else if (t.endLine > line) { break; }
        }
        if (symbol != null) { path.add(symbol); }
    }


    public synchronized ClassInfo getClass(int line, int col) {
        return contains(line, col)? findClass(line, col) : null;
    }


    private ClassInfo findClass(int line, int col) {
        ClassInfo ci = null;
        for (SymbolTable t = subTables; t != null; t = t.next) {
            if (t.contains(line, col)) {
                if ((ci = t.findClass(line, col)) != null) { break; }
            }
            else if (t.endLine > line) { break; }
        }
        if (symbol instanceof ClassInfo) {
            ci = (ClassInfo) symbol;
        }
        return ci;
    }


    public synchronized MemberInfo getMember(int line, int col) {
        return contains(line, col)? findMember(line, col) : null;
    }


    private MemberInfo findMember(int line, int col) {
        MemberInfo mi = null;
        for (SymbolTable t = subTables; t != null; t = t.next) {
            if (t.contains(line, col)) {
                if ((mi = t.findMember(line, col)) != null) { break; }
            }
            else if (t.endLine > line) { break; }
        }
        if (symbol != null) { mi = symbol; }
        return mi;
    }


    private List sv = new LinkedList();


    public synchronized MemberInfo[] getSymbols(String name, int line, int col) {
        sv.clear();

        if (contains(line, col)) { findSymbols(name, line, col, sv); }

        return (MemberInfo[]) sv.toArray(new MemberInfo[sv.size()]);
    }


    private void findSymbols(String name, int line, int col, List sv) {
        for(SymbolTable t = subTables; t != null; t = t.next) {
            if (t.contains(line, col)) {
                t.findSymbols(name, line, col, sv);
            } else if (t.endLine > line) {
                break;
            }
        }

        for (Iterator i = symbols.iterator(); i.hasNext(); ) {
            MemberInfo s = (MemberInfo) i.next();
            if (s.getName().equals(name)) { sv.add(s); }
        }
    }


    public String toString() {
        return toString("");
    }


    private String toString(String prefix) {
        String s = prefix + "{" + '\n';
        for (Iterator i = symbols.iterator(); i.hasNext(); ) {
            s += prefix + "  " + ((MemberInfo) i.next()).getLine() + '\n';
        }
        for(SymbolTable st = subTables; st != null; st = st.next) {
            s += st.toString("  " + prefix);
        }
        s += prefix + "}" + '\n';
        return s;
    }
}
