/*
 * CodeAidPopupPluginPart.java - part of the CodeAid plugin for jEdit
 * Copyright (C) 1999 Jason Ginchereau
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package codeaid.popup;

import CodeAidUtilities;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.swing.*;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

import org.gjt.sp.jedit.EBMessage;
import org.gjt.sp.jedit.EBPlugin;
import org.gjt.sp.jedit.jEdit;
import org.gjt.sp.jedit.View;
import org.gjt.sp.jedit.msg.BufferUpdate;
import org.gjt.sp.jedit.msg.PropertiesChanged;
import org.gjt.sp.jedit.msg.ViewUpdate;
import org.gjt.sp.jedit.textarea.JEditTextArea;

import org.gjt.sp.util.Log;


/**
 *
 * @author Jason Ginchereau
 * @author Andre Kaplan
**/
public class CodeAidPopupPluginPart extends EBPlugin
{
    private static DocumentHandler characterHandler = new DocumentHandler();

    private static boolean memberPopup =
        jEdit.getBooleanProperty("codeaid.automembers", true);
    private static int memberDelay =
        CodeAidUtilities.getIntegerProperty("codeaid.automembers-delay", 500);
    private static Hashtable memberTimers  = new Hashtable();
    private static Hashtable memberActions = new Hashtable();

    private static boolean methodPopup =
        jEdit.getBooleanProperty("codeaid.automethods", true);
    private static int methodDelay =
        CodeAidUtilities.getIntegerProperty("codeaid.automethods-delay", 500);
    private static Hashtable methodTimers  = new Hashtable();
    private static Hashtable methodActions = new Hashtable();


    public void start() {}


    public void stop() {}


    public void handleMessage(EBMessage message) {
        if (message instanceof BufferUpdate) {
            BufferUpdate bu = (BufferUpdate) message;
            String name = bu.getBuffer().getName().toLowerCase();

            if (name.endsWith(".java")) {
                if (bu.getWhat() == BufferUpdate.CREATED) {
                    bu.getBuffer().addDocumentListener(characterHandler);
                } else if (bu.getWhat() == BufferUpdate.CLOSED) {
                    bu.getBuffer().removeDocumentListener(characterHandler);
                }
            }
        }  else if (message instanceof ViewUpdate) {
            ViewUpdate vu = (ViewUpdate) message;

            if (vu.getWhat() == ViewUpdate.CREATED) {
                addTimersTo(vu.getView());
            } else if (vu.getWhat() == ViewUpdate.CLOSED) {
                removeTimersFrom(vu.getView());
            }
        } else if (message instanceof PropertiesChanged) {
            propertiesChanged();
        }
    }


    private static class DocumentHandler implements DocumentListener {
        public void changedUpdate(DocumentEvent e) {
            this.stopTimers();
        }

        public void insertUpdate(DocumentEvent e) {
            this.stopTimers();

            if (e.getLength() != 1) {
                return;
            }

            Document doc  = e.getDocument();
            String   text = "";
            try {
                text = doc.getText(e.getOffset(), e.getLength());
                if (!(
                           (text.equals(".") && memberPopup)
                        || (text.equals("(") && methodPopup)
                )) {
                    return;
                }

                View[] views = jEdit.getViews();

                View          view     = null;
                JEditTextArea textArea = null;
                // We look for the textArea which has the focus
                for (int i = 0; i < views.length; i++) {
                    if (views[i].getEditPane().getTextArea().hasFocus()) {
                        view     = views[i];
                        textArea = views[i].getEditPane().getTextArea();
                        break;
                    }
                }

                if (textArea == null) {
                    Log.log(Log.DEBUG, this, "No TextArea has the focus");
                    return;
                }

                Command       command = null;
                Timer         timer   = null;
                CommandAction action  = null;

                if (text.equals(".") && memberPopup) {
                    Log.log(Log.DEBUG, this, "dot typed");
                    command = new DotPopupCommand(doc, e.getOffset(), textArea);

                    if (memberDelay > 0) {
                        timer  = getMemberTimerFor(view);
                        action = getMemberActionFor(view);
                    }
                } else if (text.equals("(") && methodPopup) {
                    Log.log(Log.DEBUG, this, "bracket typed");
                    command = new BracketPopupCommand(doc, e.getOffset(), textArea);

                    if (methodDelay > 0) {
                        timer  = getMethodTimerFor(view);
                        action = getMethodActionFor(view);
                    }
                }

                if (timer == null) {
                    SwingUtilities.invokeLater(new CommandRunnable(command));
                } else {
                    Log.log(Log.DEBUG, this, "Timer running: " + timer.isRunning());

                    if (timer.isRunning()) {
                        Log.log(Log.DEBUG, this, "Timer restarted");
                        timer.stop();
                        action.setCommand(command);
                        timer.restart();
                    } else {
                        Log.log(Log.DEBUG, this, "Timer started");
                        action.setCommand(command);
                        timer.start();
                    }
                }
            } catch (BadLocationException ble) {
                Log.log(Log.DEBUG, this, ble);
            }
        }

        public void removeUpdate(DocumentEvent e) {
            this.stopTimers();
        }

        private void stopTimers() {
            for (Enumeration e = memberTimers.elements(); e.hasMoreElements(); ) {
                Timer t = (Timer) e.nextElement();

                if (t.isRunning()) {
                    t.stop();
                }
            }

            for (Enumeration e = methodTimers.elements(); e.hasMoreElements(); ) {
                Timer t = (Timer) e.nextElement();

                if (t.isRunning()) {
                    t.stop();
                }
            }
        }
   }


    private static class CommandAction implements ActionListener {
        private Command command;

        public CommandAction(Command command) {
            this.command = command;
        }

        public void actionPerformed(ActionEvent evt) {
            if (this.command != null) { this.command.execute(); }
        }

        public void setCommand(Command command) {
            this.command = command;
        }

        public Command getCommand() {
            return this.command;
        }
    }


    private static class CommandRunnable implements Runnable {
        private Command command;

        public CommandRunnable(Command command) {
            this.command = command;
        }

        public void run() {
            if (this.command != null) { this.command.execute(); }
        }
    }


    private interface Command {
        void execute();
    }


    private static class DotPopupCommand implements Command {
        private Document      document;
        private int           offset;
        private JEditTextArea textArea;


        public DotPopupCommand(Document document, int offset, JEditTextArea textArea) {
            this.document = document;
            this.offset   = offset;
            this.textArea = textArea;
        }


        public void execute() {
            synchronized (this.document) {
                PopupKeyHandler.popupKey.dotTyped(
                    this.textArea, this.document, this.offset, this.offset, ""
                );
            }
        }
    }


    private static class BracketPopupCommand implements Command {
        private Document      document;
        private int           offset;
        private JEditTextArea textArea;


        public BracketPopupCommand(Document document, int offset, JEditTextArea textArea) {
            this.document = document;
            this.offset   = offset;
            this.textArea = textArea;
        }


        public void execute() {
            synchronized (this.document) {
                PopupKeyHandler.popupKey.openParenTyped(
                    this.textArea, this.document, this.offset, ""
                );
            }
        }
    }


    private static void addTimersTo(View view) {
        CommandAction action = new CommandAction(null);
        Timer timer = new Timer(memberDelay, action);
        timer.setRepeats(false);
        memberTimers.put(view, timer);
        memberActions.put(view, action);

        action = new CommandAction(null);
        timer = new Timer(methodDelay, action);
        timer.setRepeats(false);
        methodTimers.put(view, timer);
        methodActions.put(view, action);
    }


    private static void removeTimersFrom(View view) {
        memberTimers.remove(view);
        memberActions.remove(view);

        methodTimers.remove(view);
        methodActions.remove(view);
    }


    private static Timer getMemberTimerFor(View view) {
        return (Timer) memberTimers.get(view);
    }


    private static CommandAction getMemberActionFor(View view) {
        return (CommandAction) memberActions.get(view);
    }


    private static Timer getMethodTimerFor(View view) {
        return (Timer) methodTimers.get(view);
    }


    private static CommandAction getMethodActionFor(View view) {
        return (CommandAction) methodActions.get(view);
    }


    private static void propertiesChanged() {
        memberPopup = jEdit.getBooleanProperty("codeaid.automembers", true);
        memberDelay = CodeAidUtilities.getIntegerProperty(
            "codeaid.automembers-delay", 500
        );

        for (Enumeration e = memberTimers.elements(); e.hasMoreElements(); ) {
            Timer t = (Timer) e.nextElement();

            t.setInitialDelay(memberDelay);
            t.setDelay(memberDelay);
            t.setRepeats(false);
        }

        methodPopup = jEdit.getBooleanProperty("codeaid.automethods", true);
        methodDelay = CodeAidUtilities.getIntegerProperty(
            "codeaid.automethods-delay", 500
        );

        for (Enumeration e = methodTimers.elements(); e.hasMoreElements(); ) {
            Timer t = (Timer) e.nextElement();

            t.setInitialDelay(methodDelay);
            t.setDelay(methodDelay);
            t.setRepeats(false);
        }
    }
}
