/*
 * MemberPopup.java - part of the CodeAid plugin.
 * Copyright (C) 1999 Jason Ginchereau
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package codeaid.popup;

import codeaid.info.MemberInfo;
import codeaid.info.MethodInfo;

import java.awt.Component;
import java.awt.Font;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import javax.swing.*;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

import java.util.EventObject;

// Collections API
import java.util.List;

import org.gjt.sp.jedit.EditAction;
import org.gjt.sp.jedit.View;
import org.gjt.sp.jedit.gui.InputHandler;
import org.gjt.sp.jedit.textarea.JEditTextArea;

import org.gjt.sp.util.Log;


public class MemberPopup extends CodeAidPopup {
    private JList     list;


    public MemberPopup(List listData) {
        this((Object[])listData.toArray(new Object[listData.size()]));
    }


    public MemberPopup(Object[] listData) {
        list = new JList(listData);
        list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        JScrollPane sp = new JScrollPane(list);
        sp.setBorder(null);
        add(sp);
        setVisibleRowCount(Math.min(listData.length, 5));
        list.addMouseListener(this);
        list.setCellRenderer(new MemberListCellRenderer(listData));
    }


    protected CodeAidPopup.KeyHandler getKeyHandler(InputHandler ih) {
        return new MemberPopup.KeyHandler(ih);
    }


    public void setListFont(Font f) {
        ((MemberListCellRenderer)list.getCellRenderer()).setFont(f);
    }


    public int getNameOffset() {
        return 1 + ((MemberListCellRenderer)list.getCellRenderer()).getNameOffset();
    }


    public int getVisibleRowCount() {
        return list.getVisibleRowCount();
    }


    public void setVisibleRowCount(int visibleRowCount) {
        list.setVisibleRowCount(visibleRowCount);
    }


    public void mouseClicked(MouseEvent evt) {
        useCurrentSelection(evt);
    }


    protected void updateTypedText() {
        if (typedText.length() > 0) {
            ListModel m = list.getModel();
            for (int i = 0; i < m.getSize(); i++) {
                if (m.getElementAt(i).toString().startsWith(typedText)) {
                    setSelectedIndex(i);
                    return;
                }
            }
        }
        list.getSelectionModel().clearSelection();
        if (hint != null) { hint.setVisible(false); }
    }


    private void setSelectedIndex(int index) {
        // for some reason this line makes it much faster
        list.ensureIndexIsVisible(index);

        list.setSelectedIndex(index);
        int rowCount = getVisibleRowCount();
        list.ensureIndexIsVisible(Math.max(index - (rowCount - 1) / 2, 0));
        list.ensureIndexIsVisible(Math.min(index + rowCount / 2,
                                           list.getModel().getSize()-1));

        if (hint != null) {
            Object value = list.getSelectedValue();
            String comment = (value instanceof MemberInfo ?
                              ((MemberInfo)value).getOneLineComment() : null);
            if (comment != null) {
                int i;
                while ((i = comment.indexOf("<code>")) >= 0) {
                    comment = comment.substring(0, i) + comment.substring(i+6);
                }
                while ((i = comment.indexOf("</code>")) >= 0) {
                    comment = comment.substring(0, i) + comment.substring(i+7);
                }
                comment.replace('\n', ' ');
                hint.setHint(comment);
                hint.setVisible(true);
            }
            else {
                hint.setVisible(false);
            }
        }
    }


    private void useCurrentSelection(EventObject evt) {
        Object value = list.getSelectedValue();
        if (value != null) {
            String s;
            if (value instanceof MethodInfo) {
                s = ((MemberInfo)value).getName();
                if (((MethodInfo)value).getParameterTypes().length == 0) {
                    s += "()";
                }
            }
            else if (value instanceof MemberInfo) {
                s = ((MemberInfo)value).getName();
            }
            else {
                s = value.toString();
            }

            String text = typedText;
            setVisible(false);
            JEditTextArea ta = (JEditTextArea) getInvoker();
            ta.requestFocus();
            try {
                if (text.length() > 0) {
                ta.getBuffer().remove(ta.getCaretPosition() - text.length(),
                                      text.length());
                }
                ta.getBuffer().insertString(ta.getCaretPosition(), s, null);
                if (value instanceof MethodInfo &&
                    ((MethodInfo)value).getParameterTypes().length > 0
                ) {
                    View v = EditAction.getView((Component) evt.getSource());
                    v.getInputHandler().keyTyped(new KeyEvent(ta, KeyEvent.KEY_TYPED,
                        System.currentTimeMillis(), 0, KeyEvent.VK_UNDEFINED, '(')
                    );
                }
            } catch(BadLocationException blex) { }
            getInvoker().requestFocus();
        }
    }


    protected class KeyHandler extends CodeAidPopup.KeyHandler {
        public KeyHandler(InputHandler ih) {
            super(ih);
        }


        public void keyTyped(KeyEvent evt) {
            super.keyTyped(evt);
            char c = evt.getKeyChar();
            if (c != KeyEvent.CHAR_UNDEFINED && c != '\b' && c != '\t') {
                if (Character.isJavaIdentifierPart(c)) {
                    updateTypedText();
                } else {
                    setVisible(false);
                }
            }
        }


        public void keyPressed(KeyEvent evt) {
            Log.log(Log.DEBUG, this, "keyPressed code: [" + evt.getKeyCode() + "]");
            Log.log(Log.DEBUG, this, "keyPressed source: [" + evt.getSource() + "]");

            switch (evt.getKeyCode()) {
            case KeyEvent.VK_UP:
                setSelectedIndex(Math.max(list.getSelectedIndex() - 1, 0));
                evt.consume();
                break;

            case KeyEvent.VK_DOWN:
                setSelectedIndex(Math.min(list.getSelectedIndex() + 1,
                                          list.getModel().getSize() - 1));
                evt.consume();
                break;

            case KeyEvent.VK_PAGE_UP:
                setSelectedIndex(Math.max(list.getSelectedIndex() -
                                          list.getVisibleRowCount(), 0));
                evt.consume();
                break;

            case KeyEvent.VK_PAGE_DOWN:
                setSelectedIndex(Math.min(list.getSelectedIndex() +
                                          list.getVisibleRowCount(),
                                          list.getModel().getSize() - 1));
                evt.consume();
                break;

            case KeyEvent.VK_ENTER: // Fall-through, enter or tab have the same behaviour
            case KeyEvent.VK_TAB:
                useCurrentSelection(evt);
                evt.consume();
                break;

            case KeyEvent.VK_BACK_SPACE:
                super.keyPressed(evt);
                updateTypedText();
                break;

            default:
                super.keyPressed(evt);
            }
        }
    }
}
