/*
 * PopupKeyHandler.java - part of the CodeAid plugin for jEdit
 * Copyright (C) 1999 Jason Ginchereau
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package codeaid.popup;

import codeaid.info.ClassInfo;
import codeaid.info.ClassTable;
import codeaid.info.MemberInfo;
import codeaid.info.SourceLocation;
import codeaid.info.StringComparator;

import codeaid.parser.CodeAidParserPluginPart;
import codeaid.parser.JavaParser;

import org.gjt.sp.jedit.textarea.JEditTextArea;

import java.io.EOFException;
import java.io.IOException;
import java.io.PushbackReader;

// Collections API
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;

import java.awt.Dimension;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

import java.lang.reflect.Modifier;


public final class PopupKeyHandler {
    public static PopupKeyHandler popupKey = new PopupKeyHandler();

    ClassTable  classTable;
    JavaParser  expressionParser;


    private PopupKeyHandler() {
        classTable = ClassTable.getInstance();
        expressionParser = new JavaParser();
    }


    void dotTyped(JEditTextArea ta, Document doc, int offset,
                   int showOffset, String plusText) {
        String expr = ExpressionFinder.getExpression(doc, offset + 1);
        if (expr == null) { return; }
        expr = expr.substring(0, expr.length() - 1); // remove the '.'

        JavaParser docParser = CodeAidParserPluginPart.getParser(doc);
        int line = ta.getLineOfOffset(offset);
        int col = offset - ta.getLineStartOffset(line);
        String type = expressionParser.parseExpression(expr,
            new SourceLocation(docParser.getPath(), line+1, col), docParser
        );

        ClassInfo enclosingClass = docParser.getEnclosingClass(line+1, col);

        boolean staticMembers = false;
        if (type.startsWith("*")) {
            staticMembers = true;
            type = type.substring(1);
        }

        List memberList = null;

        ClassInfo exprClass = (ClassInfo) classTable.get(type);
        if (exprClass != null) {
            memberList = getMemberList(exprClass, enclosingClass,
                                       staticMembers, null);
        }

        if (memberList == null) {
            memberList = new LinkedList();
            String packageName = expr;

            // remove any beginning new
            if (packageName.startsWith("new") &&
                Character.isWhitespace(packageName.charAt(3))
            ) {
                packageName = packageName.substring(3).trim();
            }

            // remove all spaces
            int space;
            while((space = packageName.indexOf(' ')) >= 0) {
                packageName = packageName.substring(0, space)
                            + packageName.substring(space+1);
            }

            for (Iterator i = classTable.getPackages().iterator(); i.hasNext(); ) {
                String p = (String) i.next();
                if (p.startsWith(packageName + '.')) {
                    p = p.substring(packageName.length() + 1);
                    int dot = p.indexOf('.');
                    if (dot >= 0) { p = p.substring(0, dot); }
                    memberList.add(p);
                }
            }
            if (memberList.size() > 0) {
                Collections.sort(memberList, new StringComparator());
                removeDuplicates(memberList);
            }

            List classesInPackage = classTable.getClassesInPackage(packageName);
            if (classesInPackage != null) {
                Collections.sort(classesInPackage, new StringComparator());
                memberList.addAll(classesInPackage);
            }
        }

        if (memberList != null && memberList.size() > 0) {
            line = ta.getLineOfOffset(showOffset);
            col = showOffset - ta.getLineStartOffset(line);

            MemberPopup popup = new MemberPopup(memberList);
            popup.setListFont(ta.getPainter().getFont());
            int x = ta.offsetToX(line, col + 1) - popup.getNameOffset();
            Dimension ps = popup.getPreferredSize();
            if (ps.width > 100 && ps.width > ta.getWidth() - x) {
                popup.setPreferredSize(new Dimension(ta.getWidth() - x, ps.height));
            }

            int virtualLine = ta.getBuffer().physicalToVirtual(line);
            HintPopup hint = new HintPopup(ta);
            hint.setTextLocation(virtualLine, col+1);
            hint.setLocation(Math.max(0, hint.getX() - popup.getNameOffset()),
                             hint.getY());

            popup.setHint(hint);
            popup.setStartingText(plusText);
            popup.show(ta, x, ta.lineToY(virtualLine + 1) + 5);
        }
    }


    public void showMemberSuggestionList(JEditTextArea ta) {
        String name = null;
        Document doc = ta.getBuffer();
        int offset = ta.getCaretPosition();
        int dotOffset = offset;
        try {
            int start = Math.max(0, offset - 100);
            name = doc.getText(start, offset - start);
        } catch(BadLocationException blex) {
        }

        if (name != null) {
            int dot = name.lastIndexOf(".");
            if (dot < 0) { return; }
            else { name = name.substring(dot+1); }
            dotOffset = offset - name.length() - 1;
            while (name.length() > 0 && Character.isWhitespace(name.charAt(0))) {
                name = name.substring(1);
            }
            if (name.length() > 0) {
                if (!Character.isJavaIdentifierStart(name.charAt(0))) {
                    name = null;
                } else {
                    for (int i = 1; i < name.length(); i++) {
                        if (!Character.isJavaIdentifierPart(name.charAt(i))) {
                            name = null;
                            break;
                        }
                    }
                }
            }
        }

        if (name != null) {
            dotTyped(ta, doc, dotOffset, offset-name.length()-1, name);
        }
    }


    private List getMemberList(ClassInfo ci, ClassInfo fci,
                               boolean staticMembers, String methodName) {
        List memberList = new LinkedList();
        findMemberList(ci, fci, staticMembers, memberList, methodName);
        Collections.sort(memberList);
        removeDuplicates(memberList);

        return memberList;
    }


    // assumes list is already sorted
    private static void removeDuplicates(List list) {
        if (list.size() > 0) {
            Iterator i = list.iterator();
            Object prev = i.next();
            while (i.hasNext()) {
                Object o = i.next();
                if (o.equals(prev)) { i.remove(); }
                prev = o;
            }
        }
    }


    private void findMemberList(ClassInfo ci, ClassInfo fci,
                                boolean staticMembers, List memberList,
                                String methodName) {
        if (methodName == null) {
            for (Iterator i = ci.getClasses().iterator(); i.hasNext(); ) {
                MemberInfo mi = (MemberInfo)
                    classTable.get(ci.getFullName() + '$' + (String) i.next());
                if (mi != null &&
                    classTable.isAccessAllowed(fci, mi, staticMembers, false)) {
                    memberList.add(mi);
                }
            }
            for (Iterator i = ci.getFields().iterator(); i.hasNext(); ) {
                MemberInfo mi = (MemberInfo) i.next();
                if (classTable.isAccessAllowed(fci, mi, staticMembers, false)) {
                    memberList.add(mi);
                }
            }
        }

        for (Iterator i = ci.getMethods().iterator(); i.hasNext(); ) {
            MemberInfo mi = (MemberInfo) i.next();
            if(methodName == null || methodName.equals(mi.getName())) {
                if(classTable.isAccessAllowed(fci, mi, staticMembers, false)) {
                    memberList.add(mi);
                }
            }
        }

        String[] ifs = ci.getInterfaces();
        String sc = ci.getSuperclass();
        for (int i = 0; i < ifs.length; i++) {
            ci = classTable.get(ifs[i]);
            if (ci != null) {
                findMemberList(ci, fci, staticMembers, memberList, methodName);
            }
        }
        if (sc != null) {
            ci = classTable.get(sc);
            if (ci != null) {
                findMemberList(ci, fci, staticMembers, memberList, methodName);
            }
        }
    }


    void openParenTyped(JEditTextArea ta, Document doc, int offset,
                              String plusText) {
        int showOffset = offset;
        String name = null;
        List memberList = null;
        try {
            int start = Math.max(0, offset - 100);
            name = doc.getText(start, offset - start + 1);
        } catch(BadLocationException blex) {
        }

        if (name != null && name.endsWith("(")) {
            name = name.substring(0, name.length()-1);
            int newIndex = ExpressionFinder.findConstructorNew(name);
            if (newIndex >= 0) {
                name = name.substring(newIndex + 3).trim();
                String newExpr = "new " + name + "()";

                JavaParser docParser = CodeAidParserPluginPart.getParser(doc);
                int line = ta.getLineOfOffset(offset);
                int col = offset - ta.getLineStartOffset(line);
                String type = expressionParser.parseExpression(newExpr,
                    new SourceLocation(docParser.getPath(), line+1, col), docParser);

                ClassInfo ci = classTable.get(type);
                if (ci == null) { return; }

                memberList = new LinkedList(ci.getConstructors());
            }
            else {
                String expr;
                int dotOffset;
                int i = name.length() - 1;
                while (i >= 0 && Character.isWhitespace(name.charAt(i))) { i--; }
                while (i >= 0 && Character.isJavaIdentifierPart(name.charAt(i))) { i--; }
                while (i >= 0 && Character.isWhitespace(name.charAt(i))) { i--; }
                if (name.charAt(i) == '.') {
                    name = name.substring(i+1);
                    dotOffset = offset - name.length() - 1;
                    name = name.trim();
                    expr = ExpressionFinder.getExpression(doc, dotOffset + 1);
                }
                else {
                    name = name.substring(i+1);
                    dotOffset = offset - name.length() - 1;
                    name = name.trim();
                    expr = "this.";
                }

                if (expr == null) { return; }
                expr = expr.substring(0, expr.length() - 1); // remove the '.'

                JavaParser docParser = CodeAidParserPluginPart.getParser(doc);
                int line = ta.getLineOfOffset(dotOffset);
                int col = dotOffset - ta.getLineStartOffset(line);
                String type = expressionParser.parseExpression(expr,
                    new SourceLocation(docParser.getPath(), line+1, col), docParser);

                ClassInfo enclosingClass = docParser.getEnclosingClass(line+1, col);

                boolean staticMembers = false;
                if (type.startsWith("*")) {
                    staticMembers = true;
                    type = type.substring(1);
                }

                ClassInfo ci = classTable.get(type);
                if (ci == null) { return; }

                memberList = getMemberList(ci, enclosingClass, staticMembers, name);
            }
        }

        if (memberList != null && memberList.size() > 0) {
            int line = ta.getLineOfOffset(showOffset);
            int col = showOffset - ta.getLineStartOffset(line);

            MethodPopup popup = new MethodPopup(memberList);
            popup.setListFont(ta.getPainter().getFont());
            int x = ta.offsetToX(line, col+1)-popup.getNameOffset();
            Dimension ps = popup.getPreferredSize();
            if (ps.width > 100 && ps.width > ta.getWidth() - x) {
                popup.setPreferredSize(new Dimension(ta.getWidth() - x, ps.height));
            }

            int virtualLine = ta.getBuffer().physicalToVirtual(line);
            HintPopup hint = new HintPopup(ta);
            hint.setTextLocation(virtualLine, col+1);
            hint.setLocation(Math.max(0, hint.getX() - popup.getNameOffset()),
                             hint.getY());

            popup.setHint(hint);
            popup.setStartingText(plusText);
            popup.show(ta, x, ta.lineToY(virtualLine + 1) + 5);
        }
    }


    public void showMethodHelp(JEditTextArea ta) {
        String s = null;
        Document doc = ta.getBuffer();
        int offset = ta.getCaretPosition();
        int parenOffset = offset;
        try {
        int start = Math.max(0, offset - 100);
        s = doc.getText(start, offset - start);
        } catch(BadLocationException blex) {}

        if (s != null) {
            int paren = -1;
            int parenCount = 0;
            for (int i = s.length() - 1; i >= 0; i--) {
                switch (s.charAt(i)) {
                case ')': parenCount++; break;
                case '(': parenCount--; break;
                }
                if (parenCount < 0) {
                    paren = i;
                    break;
                }
            }
            if (paren < 0) { return; }

            s = s.substring(paren+1);
            parenOffset = offset - s.length() - 1;
            openParenTyped(ta, doc, parenOffset, s);
        }
    }


    private static class ExpressionFinder {
        static class Token {
            static final String[] names =
                { "ERROR", "DOT", "STRLIT", "IDENT", "PAREN", "BRACKET", "NEW"};

            static final int ERROR   = 0;
            static final int DOT     = 1;
            static final int STRLIT  = 2;
            static final int IDENT   = 3;
            static final int PAREN   = 4;
            static final int BRACKET = 5;
            static final int NEW     = 6;

            public int start;
            public int end;
            public int type;
        }


        static class States {
            static final int ERROR   = 0; // accept last more-accept, if any
            static final int START   = 1;
            static final int DOT     = 2;
            static final int BRACKET = 3;
            static final int PAREN   = 4;
            static final int MORE    = 5; // accept but maybe get more
            static final int ACCEPT  = 6; // accept as of this token

            static final int[][] table =
                         /*  ERROR,   DOT,STRLIT, IDENT, PAREN, BRACKET, NEW */
                /* ERROR */{{    0,     0,     0,     0,     0,     0,     0 },
                /* START */ {    0,   DOT,     0,     0,     0,     0,     0 },
                /*  DOT  */ {    0,     0,ACCEPT,  MORE, PAREN, BRACKET,   0 },
                /*BRACKET*/ {    0,     0,     0,  MORE, PAREN,     0,     0 },
                /* PAREN */ {    0,     0,     0,  MORE,     0,     0,     0 },
                /* MORE  */ {    0,   DOT,     0,     0,     0,     0,ACCEPT }};
        }


        /**
         * Gets the entire expression that ends at offset.
         *
         * This method must be run inside a synchronization lock on the document.
        **/
        public static String getExpression(Document doc, int endOffset) {
            String expr = "";
            int lastAccept = endOffset;
            int state = States.START;
            int prevState = -1, prevState2 = -1;
            Token token = new Token();

            try {
                ReverseDocumentReader rdr = new ReverseDocumentReader(doc, endOffset);
                PushbackReader r = new PushbackReader(rdr);

                while (state != States.ACCEPT && state != States.ERROR) {
                    nextToken(r, rdr, token);
                    //System.err.println(Token.names[token.type]);

                    prevState2 = prevState;
                    prevState = state;
                    state = States.table[state][token.type];

                    if (state == States.ACCEPT || state == States.MORE) {
                        lastAccept = token.start;
                    }

                    // special case for just (expr).
                    if (   (state      == States.PAREN)
                        && (prevState  == States.DOT)
                        && (prevState2 == States.START)
                    ) {
                        lastAccept = token.start;
                    }
                }

                if (lastAccept < endOffset) {
                    return doc.getText(lastAccept, endOffset - lastAccept);
                }
            } catch(BadLocationException blex) {
                // System.err.println(blex);
            }

            return null;
        }


        private static void nextToken(PushbackReader r, ReverseDocumentReader rdr,
                                      Token t) {
            try {
                // peek
                int c;
                do { c = r.read(); }
                while (Character.isWhitespace((char)c));
                if (c >= 0) { r.unread(c); }

                t.type = Token.ERROR;
                t.start = t.end = rdr.getOffset();

                if (Character.isJavaIdentifierPart((char)c)) {
                    String ident = ident(r);
                    if (ident != null) {
                        if (ident.equals("new")) {
                            t.type = Token.NEW;
                        } else {
                            t.type = Token.IDENT;
                        }
                        t.start = rdr.getOffset() + 1;
                        return;
                    }
                } else {
                    switch (c) {
                    case '"':
                    if (strlit(r)) { t.type = Token.STRLIT; }
                        t.start = rdr.getOffset() + 1;
                        break;

                    case ')':
                    if (bracket(r)) { t.type = Token.PAREN; }
                        t.start = rdr.getOffset();
                        break;

                    case ']':
                    case '}':
                    if (bracket(r)) { t.type = Token.BRACKET; }
                        t.start = rdr.getOffset();
                        break;

                    case '.':
                    if (r.read() == '.') { t.type = Token.DOT; }
                        break;

                    default:
                    }
                }
            } catch(IOException ioex) {
                // System.err.println(ioex);
            }
        }


        private static String ident(PushbackReader r) throws IOException {
            String ident = "";
            int c;
            while (true) {
                c = r.read();
                if (c < 0) { throw new EOFException(); }
                if (!Character.isJavaIdentifierPart((char)c)) { break; }
                ident = "" + ((char)c) + ident; // reverse append
            }
            r.unread(c);
            if (ident.length() > 0 &&
                Character.isJavaIdentifierStart(ident.charAt(0))
            ) {
                if(!ident.equals("while") &&
                   !ident.equals("for") &&
                   !ident.equals("do") &&
                   !ident.equals("if") &&
                   !ident.equals("else") &&
                   !ident.equals("synchronized") &&
                   !ident.equals("try") &&
                   !ident.equals("catch") &&
                   !ident.equals("finally")
                ) {
                    return ident;
                }
            }
            return null;
        }


        private static boolean strlit(PushbackReader r) throws IOException {
            int c = r.read();
            if (c < 0) { throw new EOFException(); }
            if (c != '"') {
                r.unread(c);
                return false;
            }
            while (true) {
                c = 0;
                while (c != '"') {
                    c = r.read();
                    if (c < 0) { throw new EOFException(); }
                }
                c = r.read();
                if (c < 0) { throw new EOFException(); }
                if (c != '\\') {
                    r.unread(c);
                    return true;
                }
            }
        }


        private static boolean bracket(PushbackReader r) throws IOException {
            int b = r.read();
            if (b < 0) { throw new EOFException(); }
            if (b != ']' && b != '}' && b != ')') { return false; }
            r.unread(b);

            int c = 0, prev = 0, prev2 = 0;

            boolean cquote = false;
            boolean squote = false;
            boolean comment = false;
            int bracketLevel = 0;
            int parenLevel = 0;
            int braceLevel = 0;

            while(true) {
                prev2 = prev;
                prev = c;
                c = r.read();
                if (c < 0) { throw new EOFException(); }

                switch (c) {
                case '\'':
                    if (!comment && !squote) { cquote = !cquote; }
                    break;

                case  '"':
                    if (!comment && !cquote) { squote = !squote; }
                    break;

                case  ']':
                    if (!comment && !squote && !cquote) { bracketLevel++; }
                    break;

                case  '[':
                    if (!comment && !squote && !cquote) { bracketLevel--; }
                    break;

                case  ')':
                    if (!comment && !squote && !cquote) { parenLevel++; }
                    break;

                case  '(':
                    if (!comment && !squote && !cquote) { parenLevel--; }
                    break;

                case  '}':
                    if (!comment && !squote && !cquote) { braceLevel++; }
                    break;

                case  '{':
                    if (!comment && !squote && !cquote) { braceLevel--; }
                    break;

                case  '/':
                    if (!squote && !cquote && prev == '*') { comment = false; }
                    break;

                case  '*':
                    if (!squote && !cquote && prev == '/') { comment = true; }
                    break;

                case '\\':
                    if(!comment) {
                        switch(prev) {
                        case '\'':
                            if (!squote) { cquote = !cquote; }
                            break;

                        case  '"':
                            if (!cquote) { squote = !squote; }
                            break;
                        }
                    }
                    break;
                }

                if (bracketLevel < 0 || parenLevel < 0 || braceLevel < 0) {
                    return false;
                }
                if (bracketLevel == 0 && parenLevel == 0 && braceLevel == 0) {
                    return true;
                }
            }
        }


        public static int findConstructorNew(String s) {
            int i;
            for (i = s.length() - 1; i >= 0; i--) {
                char c = s.charAt(i);
                if (!(Character.isWhitespace(c) ||
                      Character.isJavaIdentifierPart(c) || c == '.')) { break; }
                if (i > 3 && c == 'w' && Character.isWhitespace(s.charAt(i+1)) &&
                    !Character.isJavaIdentifierPart(s.charAt(i-3)) &&
                    s.regionMatches(i-2, "new", 0, 3)) { return i-2; }
            }
            return -1;
        }
    }
}
