/*
 * CDBBuilder.java - part of the CodeAid plugin.
 * Copyright (C) 1999 Jason Ginchereau
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package codeaid.tools;

import codeaid.info.ClassTable;

import java.io.File;

// Collections API
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;


/**
 * @author Jason Ginchereau
 */
public class CDBBuilder implements Runnable {
    private static void usage() {
        System.err.println(
  "\n"
+ "Usage: cdbbuild <options>\n"
+ "\n"
+ "Parses source files to a database for use with CodeAid.\n"
+ "\n"
+ "-f --fullcomments <yes|NO>      store full javadoc comments (> 1 sentence)\n"
+ "-p --privateinfo  <yes|NO>      store private and package-private info    \n"
+ "-s --source  <dirname>          parse .java files in directory tree       \n"
+ "-s --source  <zipfile>[!<path>] parse .java files in zip/jar              \n"
+ "-i --include <filename.cdb>     include other .cdb in output              \n"
+ "-r --read    <filename.cdb>     read other .cdb but don't include in output\n"
+ "-o --output  <filename.cdb>     output new database to .cdb file          \n"
+ "\n"
+ "Example: cdbbuild -s src.jar!src -o java.cdb\n"
+ "Example: cdbbuild -r java.cdb -s \"C:/Program Files/jEdit\" -o jedit.cdb\n"
+ "\n"
+ "Commands are executed in the order they are parsed, which means information\n"
+ "from later sources will overwrite earlier information, and also output\n"
+ "should usually be the last command.\n"
+ "\n"
        );
    }


    public static void main(String[] args) {
        CDBBuilder c;
        try {
            c = new CDBBuilder(args);
        } catch(IllegalArgumentException iaex) {
            if(iaex.getMessage() != null && iaex.getMessage().length() > 0) {
                System.err.println(iaex.getMessage());
                System.err.println();
            }
            usage();
            return;
        }

        c.run();
    }


    private List operations;
    private ClassTable pct;
    private ClassTable ct;
    private boolean finished;
    private boolean oneLineComments = true;
    private boolean privateInfo = false;


    public CDBBuilder(String[] args) throws IllegalArgumentException {
        if (args.length == 0) { throw new IllegalArgumentException(); }

        // fix a strange bug
        for (int i = 0; i < args.length; i++) { args[i] = args[i].trim(); }

        finished = false;
        pct = new ClassTable();
        ct = new ClassTable(pct);
        operations = new LinkedList();
        for (int i = 0; i < args.length; i++) {

            // fix another strange bug
            if (args[i].length() == 0) { continue; }

            if (i + 1 < args.length) {
                if (args[i+1].startsWith("-")) {
                    throw new IllegalArgumentException("No value specified for " +
                    args[i]);
                }
                else if (args[i].equals("-f") || args[i].equals("--fullcomments")) {
                    oneLineComments = !parseBoolean(args[++i]);
                }
                else if (args[i].equals("-p") || args[i].equals("--privateinfo")) {
                    privateInfo = parseBoolean(args[++i]);
                }
                else if (args[i].equals("-s") || args[i].equals("--source"  )) {
                    operations.add(new SourceReader(args[++i], ct,
                    oneLineComments, privateInfo));
                }
                else if (args[i].equals("-c") || args[i].equals("--class")) {
                    operations.add(new ClassReader(args[++i], ct));
                }
                //else if (args[i].equals("-d" ) || args[i].equals("--doc"  )) {
                    //  operations.add(new DocReader(args[++i], ct));
                //}
                else if (args[i].equals("-r") || args[i].equals("--read" )) {
                    operations.add(new DBReader(args[++i], pct));
                }
                else if (args[i].equals("-i") || args[i].equals("--include" )) {
                    operations.add(new DBReader(args[++i], ct));
                }
                else if (args[i].equals("-o") || args[i].equals("--output"  )) {
                    operations.add(new DBWriter(args[++i], ct));
                }
                /*else if (args[i].equals("-x") || args[i].equals("--xml")) {
                    operations.add(new XMLWriter(args[++i], ct));
                }*/ else {
                    throw new IllegalArgumentException(
                        "Illegal option: " + args[i]);
                }
            } else {
                throw new IllegalArgumentException();
            }
        }
    }


    public CDBBuilder(String baseName, CDBEntry[] cdbEntries) {
        finished = false;
        pct = new ClassTable();
        ct = new ClassTable(pct);
        String pathPrefix = ((baseName == null || baseName.length() == 0)
            ? ""
            : baseName
        );
        if (!pathPrefix.endsWith(File.separator)) {
            pathPrefix += File.separator;
        }

        operations = new LinkedList();
        for (int i = 0; i < cdbEntries.length; i++) {
            CDBEntry cdbEntry = cdbEntries[i];
            operations.add(new SourceReader(cdbEntry.lib, ct, !cdbEntry.fullComments,
            cdbEntry.privateInfo));
            operations.add(new DBWriter(pathPrefix + cdbEntry.name, ct));
        }
    }


    private static boolean parseBoolean(String b) {
        b = b.toLowerCase();
        return (
               b.startsWith("1")
            || b.startsWith("t")
            || b.startsWith("y")
            || b.startsWith("on")
        );
    }


    public void run() {
        if (!finished) {
            for (Iterator i = operations.iterator(); i.hasNext(); ) {
                ((Runnable) i.next()).run();
            }
        }
        finished = true;
    }


    public boolean isFinished() {
        return finished;
    }


    public ClassTable getClassTable() {
        return ct;
    }
}

