/*
 * ClassInfoReader.java
 * Copyright (C) 1999-2001 Jason Ginchereau, Andre Kaplan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package codeaid.tools;

import codeaid.info.ClassTable;

import java.io.IOException;
import java.io.File;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.InputStream;

import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;


abstract class ClassInfoReader implements Runnable {
    private ClassTable ct;
    private File       file = null;
    private String     path = null;
    private int        numPasses;


    protected ClassInfoReader(String filename, ClassTable ct, int numPasses) {
        int i = filename.lastIndexOf("!");
        if (i >= 0) {
            path = filename.substring(i+1);
            if (path.startsWith("/")) { path = path.substring(1); }
            if (!path.endsWith("/")) { path += '/'; }
            filename = filename.substring(0, i);
            if (filename.endsWith("\\")) {
                filename = filename.substring(0, filename.length()-1);
            }
        }
        file = new File(filename);
        this.ct = ct;
        if (!file.exists()) {
            throw new IllegalArgumentException("\"" + file + "\" does not exist.");
        }
        if (!file.isDirectory()) {
            if (!(file.getPath().toLowerCase().endsWith(".zip") ||
                file.getPath().toLowerCase().endsWith(".jar"))
            ) {
                throw new IllegalArgumentException("\"" + file + "\""
                    + " is not a directory, .zip or .jar file."
                );
            }
        }
        this.numPasses = numPasses;
    }


    public final void run() {
        if (file.isDirectory()) {
            for (int n = 0; n < numPasses; n++) {
                processDir(file, "", n);
            }
        } else {
            for (int n = 0; n < numPasses; n++) {
                processZip(file, path, n);
            }
        }
    }


    private final void processDir(File dir, String pathPrefix, int n) {
        try {
            //File[] f = dir.listFiles();
            String[] fileList = dir.list();
            for (int i = 0; i < fileList.length; i++) {
                File f = new File(dir, fileList[i]);
                if (f.isDirectory()) {
                    processDir(f, pathPrefix + f.getName() + '/', n);
                } else {
                    process(new BufferedInputStream(new FileInputStream(f)),
                        pathPrefix + f.getName(), ct, n);
                }
            }
        } catch (IOException ioex) {
            ioex.printStackTrace(System.err);
        }
    }


    private final void processZip(File file, String prefixPath, int n) {
        try {
            ZipInputStream zipIn = new ZipInputStream(new BufferedInputStream(
                new FileInputStream(file)));
            ZipEntry entry;
            while ((entry = zipIn.getNextEntry()) != null) {
                String entryPath = entry.getName();
                if (prefixPath != null) {
                    if (!entryPath.startsWith(prefixPath)) {
                        zipIn.closeEntry();
                        continue;
                    }
                    entryPath = entryPath.substring(prefixPath.length());
                }
                process(new BufferedInputStream(zipIn), entryPath, ct, n);
                zipIn.closeEntry();
            }
            zipIn.close();
        } catch (IOException ioex) {
            ioex.printStackTrace(System.err);
        }
    }


    protected abstract void process(InputStream in, String path,
        ClassTable ct, int n) throws IOException;
}

